Enum OrbitalElement
pub enum OrbitalElement {
Show 53 variants
Altitude,
AoL,
AoP,
ApoapsisRadius,
ApoapsisAltitude,
BrouwerMeanShortSemiMajorAxis,
BrouwerMeanShortEccentricity,
BrouwerMeanShortInclination,
BrouwerMeanShortRAAN,
BrouwerMeanShortAoP,
BrouwerMeanShortMeanAnomaly,
C3,
Declination,
EccentricAnomaly,
Eccentricity,
EquinoctialH,
EquinoctialK,
EquinoctialP,
EquinoctialQ,
EquinoctialLambda,
Energy,
FlightPathAngle,
Height,
Latitude,
Longitude,
Hmag,
HX,
HY,
HZ,
HyperbolicAnomaly,
Inclination,
MeanAnomaly,
PeriapsisRadius,
PeriapsisAltitude,
Period,
RightAscension,
RAAN,
Rmag,
SemiParameter,
SemiMajorAxis,
SemiMajorAxisAltitude,
SemiMinorAxis,
TrueAnomaly,
TrueLongitude,
VelocityDeclination,
Vmag,
X,
Y,
Z,
VX,
VY,
VZ,
Custom,
}Expand description
Orbital element defines all of the supported orbital elements in ANISE, which are all built from a State.
Variants
Altitude
Altitude (km)
AoL
Argument of Latitude (deg)
AoP
Argument of Periapse (deg)
ApoapsisRadius
Radius of apoapsis (km)
ApoapsisAltitude
Altitude of apoapsis (km)
BrouwerMeanShortSemiMajorAxis
Brouwer Mean Short Semi major axis (km)
BrouwerMeanShortEccentricity
Brouwer Mean Short Eccentricity
BrouwerMeanShortInclination
Brouwer Mean Short Inclination (deg)
BrouwerMeanShortRAAN
Brouwer Mean Short RAAN (deg)
BrouwerMeanShortAoP
Brouwer Mean Short Argument of Periapse (deg)
BrouwerMeanShortMeanAnomaly
Brouwer Mean Short Mean Anomaly (deg)
C3
C_3 in (km/s)^2
Declination
Declination (deg) (also called elevation if in a body fixed frame)
EccentricAnomaly
Eccentric anomaly (deg)
Eccentricity
Eccentricity (no unit)
EquinoctialH
Equinoctial H element
EquinoctialK
Equinoctial K element
EquinoctialP
Equinoctial P element
EquinoctialQ
Equinoctial Q element
EquinoctialLambda
Equinoctial mean longitude Lambda (deg)
Energy
Specific energy
FlightPathAngle
Flight path angle (deg)
Height
Geodetic height (km)
Latitude
Geodetic latitude (deg)
Longitude
Geodetic longitude (deg)
Hmag
Orbital momentum
HX
X component of the orbital momentum vector
HY
Y component of the orbital momentum vector
HZ
Z component of the orbital momentum vector
HyperbolicAnomaly
Hyperbolic anomaly (deg), only valid for hyperbolic orbits
Inclination
Inclination (deg)
MeanAnomaly
Mean anomaly (deg)
PeriapsisRadius
Radius of periapse (km)
PeriapsisAltitude
Altitude of periapse (km)
Period
Orbital period (s)
RightAscension
Right ascension (deg)
RAAN
Right ascension of the ascending node (deg)
Rmag
Norm of the radius vector
SemiParameter
Semi parameter (km)
SemiMajorAxis
Semi major axis (km)
SemiMajorAxisAltitude
Semi major axis altitude (km)
SemiMinorAxis
Semi minor axis (km)
TrueAnomaly
True anomaly
TrueLongitude
True longitude
VelocityDeclination
Velocity declination (deg)
Vmag
Norm of the velocity vector (km/s)
X
X component of the radius (km)
Y
Y component of the radius (km)
Z
Z component of the radius (km)
VX
X component of the velocity (km/s)
VY
Y component of the velocity (km/s)
VZ
Z component of the velocity (km/s)
Custom
Implementations
impl OrbitalElement
impl OrbitalElement
impl OrbitalElement
impl OrbitalElement
pub fn py_evaluate(&self, orbit: CartesianState) -> Result<f64, PyErr>
pub fn py_evaluate(&self, orbit: CartesianState) -> Result<f64, PyErr>
Evaluate the orbital element enum variant for the provided orbit
:type orbit: Orbit :rtype: float