TimeScale

Enum TimeScale 

#[non_exhaustive]
pub enum TimeScale {
Show 13 variants TAI, TT, ET, TDB, UTC, GPST, GST, BDT, QZSST, TCG, TCB, TL, TCL,
}
Enum of the different time systems available

Variants (Non-exhaustive)

This enum is marked as non-exhaustive
Non-exhaustive enums could have additional variants added in future. Therefore, when matching against variants of non-exhaustive enums, an extra wildcard arm must be added to account for any future variants.


TAI

TAI is the representation of an Epoch internally



TT

Terrestrial Time (TT) (previously called Terrestrial Dynamical Time (TDT))



ET

Ephemeris Time as defined by SPICE (slightly different from true TDB)



TDB

Dynamic Barycentric Time (TDB) (higher fidelity SPICE ephemeris time)



UTC

Universal Coordinated Time



GPST

GPS Time scale whose reference epoch is UTC midnight between 05 January and 06 January 1980; cf. https://gssc.esa.int/navipedia/index.php/Time_References_in_GNSS#GPS_Time_.28GPST.29. |UTC - TAI| = 19 Leap Seconds on that day.



GST

Galileo Time scale



BDT

BeiDou Time scale



QZSST

QZSS Time scale has the same properties as GPST but with dedicated clocks



TCG

Geocentric Coordinate Time



TCB

Barycentric Coordinate Time



TL

Experimental Lunar Time, option (iii) from the Lunar Reference Timescale paper, A Bourgoin*, P Defraigne and F Meynadier

TL is defined as a linear scaling of TCL such that TL has no secular drift with respect to TT. Since this implementation omits the bounded periodic TCL-TT terms, TL is equivalent to TT after the common 1977 reference epoch.



TCL

Experimental mean Lunar Coordinate Time of Lunar reference timescale, A Bourgoin*, P Defraigne and F Meynadier

This is not a full IAU-quality TCL realization. It models only the conventional secular mean rate between TCL and TT:

    d(TCL - TT) / dTT ≈ 6.8e-10

The bounded periodic TCL-TT terms and ephemeris-dependent relativistic integral are intentionally omitted

Implementations



impl TimeScale

pub const fn is_gnss(&self) -> bool

Returns true if Self is based off a GNSS constellation

pub const fn reference_epoch(self) -> Epoch

Returns this time scale’s reference epoch: Time Scale initialization date, expressed as an Epoch in TAI



impl TimeScale

pub const fn uses_leap_seconds(&self) -> bool

Returns true if self takes leap seconds into account :rtype: bool